//
// Copyright (c) 2007, Mats Engstrom, mats.engstrom@gmail.com
// This code is a part of an Instructable published under the
// Creative Commons Attribution Non-commercial Share Alike (by-nc-sa)
//
// The Instructable can be found at 
// http://www.instructables.com/member/matseng/
//
#define F_CPU 4000000UL

#include <inttypes.h>
#include <string.h>
#include <stdlib.h>
#include <util/delay.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>

/*
        Tiny2313
        +--------------------------------------------+
  RESET |  1 PA2/-reset                       VCC 20 | +5 VOLT
      A |  2 PD0/rxd            int7/scl/usck/PB7 19 | CC3 / SCK
      B |  3 PD1/txd             int6/do/miso/PB6 18 | CC2 / MISO 
      M |  4 PA1/xtal2           int5/di/mosi/PB5 17 | CC1 / MOSI 
      C |  5 PA0/xtal1              int4/oc1b/PB4 16 | K
     G1 |  6 PD2/int0/ckout         int3/oc1a/PB3 15 | J
      D |  7 PD3/int1               int2/oc0a/PB2 14 | I
      E |  8 PD4/t0                 int1/ain1/PB1 13 | H
      F |  9 PD5/t1/oc0b            int0/ain0/PB0 12 | G2
    GND | 10 GND                          icp/PD6 11 | L 
        +--------------------------------------------+
*/ 


// The number of messages to randomly select from
#define MESSAGE_COUNT 8


// The messages
char msg1[] PROGMEM = "I LOVE YOU";
char msg2[] PROGMEM = "YOU ARE BEAUTIFUL";
char msg3[] PROGMEM = "YOU ARE THE SUNSHINE IN MY LIFE";
char msg4[] PROGMEM = "HUGS AND KISSES";
char msg5[] PROGMEM = "YOU ARE THE BEST";
char msg6[] PROGMEM = "I ALWAYS THINK OF YOU";
char msg7[] PROGMEM = "I WILL ALWAYS LOVE YOU";
char msg8[] PROGMEM = "YOU ARE MY ENTIRE LIFE";

// Build an array of the messages
PGM_P messages[MESSAGE_COUNT] PROGMEM = 
{
	msg1,	msg2,	msg3,	msg4,
	msg5,	msg6,	msg7,	msg8
};


// Storage for the random seed in the nonvolatile
// EEPROM-area. Allocate 32 bits for the seed
uint16_t EEMEM eeprom_seed; 
uint16_t EEMEM eeprom_seed_2; 


// The number of digits in the display
#define DIGIT_COUNT 8

// The array of bits (segments) to be set on each of the 8 digts
volatile uint16_t segs[DIGIT_COUNT];


// Mappings of bits to segments for the character map
#define 	A	0x0001
#define 	B	0x0002
#define 	C	0x0004
#define 	D	0x0008
#define 	E	0x0010
#define 	F	0x0020
#define 	G1	0x0040
#define 	G2	0x0080
#define 	H 	0x0100
#define 	I 	0x0200
#define 	J 	0x0400
#define 	K 	0x0800
#define 	L 	0x1000
#define 	M 	0x2000

// Build the character map
const uint16_t charactermap[64] PROGMEM = {
	0,					// space
	D+I,				// exclamation point
	B+F,				// double quote				
	B+C+D+G1+G2+I+L,	// hash
	A+C+D+F+G1+G2+I+L,	// dollar
	C+F+G1+G2+H+J+K+M,	// percent
	A+C+D+H+J+K+M,		// ampersand
	B,					// single quote
	J+K,				// left paranthesis
	H+M,				// right paranthesis
	G1+G2+H+I+J+K+L+M,	// asterisk
	G1+G2+I+L,			// plus
	M,					// comma
	G1+G2,				// minus
	E,					// dot
	J+M,				// slash
	A+B+C+D+E+F+J+M,	// zero
	I+L,				// one
	A+B+D+E+G1+G2,		// two
	A+B+C+D+G1+G2,		// three
	B+C+F+G1+G2,		// four
	A+D+F+G1+K,			// five
	A+C+D+E+F+G1+G2,	// six
	A+B+C,				// seven
	A+B+C+D+E+F+G1+G2,	// eight
	A+B+C+D+F+G1+G2,	// nine
	A+D,				// colon
	A+M,				// semi colon
	D+M,				// less than
	D+G1+G2,			// equal sign
	D+K,				// greater than
	A+B+G2+L,			// question mark
	A+B+D+E+F+I+G2,		// @
	A+B+C+E+F+G1+G2,	// a
	A+D+E+F+G1+J+K,		// b
	A+D+E+F,			// c
	A+B+C+D+I+L,		// d
	A+D+E+F+G1,			// e
	A+E+F+G1,			// f
	A+C+D+E+F+G2,		// g
	B+C+E+F+G1+G2,		// h
	I+L,				// i
	B+C+D+E,			// j
	E+F+G1+J+K,			// k
	D+E+F,				// l
	B+C+E+F+H+J,		// m
	B+C+E+F+H+K,		// n
	A+B+C+D+E+F,		// o
	A+B+E+F+G1+G2,		// p
	A+B+C+D+E+F+K,		// q
	A+B+E+F+G1+G2+K,	// r
	A+C+D+F+G1+G2,		// s
	A+I+L,				// t
	B+C+D+E+F,			// u
	E+F+J+M,			// v
	B+C+E+F+K+M,		// w
	H+J+K+M,			// x
	H+J+L,				// y
	A+D+J+M,			// z
	G2+J+K,				// left bracket
	H+K,				// backslash
	G1+H+M,				// right bracket
	A+B+J,				// caret
	D					// underscore
};




//
// A generic delay subroutine
//
void Delay_ms(int cnt) {
	while (cnt-->0) {
		_delay_ms(1);
	}
}


//
// This is the Interrupt Service Routine that need to get called 
// often enough to not show any visible flickering on the displays.
//
ISR (SIG_OVERFLOW0) { 
	static uint8_t currentdigit=0;	// Keeps track of the digit to be shown
	uint8_t	pa;				// Variable to build the PORTA-data in
	uint8_t	pb;				// Variable to build the PORTB-data in
	uint8_t	pd;				// Variable to build the PORTD-data in

	TCNT0 = 128;         	// Reset TCNT0 halfway to get a higher freq

		
	// Clear the variables so we can bitwise-OR in new values to them
	pa = 0;
	pb = 0;
	pd = 0;

	// Set the bits in the port-varaibles according to the bits 
	// in the segs-array.
	if (segs[currentdigit] & A)  pd |= 0x01;
	if (segs[currentdigit] & B)  pd |= 0x02;
	if (segs[currentdigit] & C)  pa |= 0x01;
	if (segs[currentdigit] & D)  pd |= 0x08;
	if (segs[currentdigit] & E)  pd |= 0x10;
	if (segs[currentdigit] & F)  pd |= 0x20;
	if (segs[currentdigit] & G1) pd |= 0x04;
	if (segs[currentdigit] & G2) pb |= 0x01;
	if (segs[currentdigit] & H)  pb |= 0x02;
	if (segs[currentdigit] & I)  pb |= 0x08;
	if (segs[currentdigit] & J)  pb |= 0x04;
	if (segs[currentdigit] & K)  pb |= 0x10;
	if (segs[currentdigit] & L)  pd |= 0x40;
	if (segs[currentdigit] & M)  pa |= 0x02;

	// PORTB bits 5-7 control the digits selector
	pb |= (currentdigit<<5);

	// Begin by turning of all old segments to reduce
	// possible glitching. Probably not neccessary...
	PORTD=0;
	PORTA=0;
	PORTB=0;

	// Now light up the new segments, but begin with PORTB as
	// it controls the digit
	PORTB = pb;
	PORTA = pa;
	PORTD = pd;

	// Increment the digitcounter so it gets displayed at next
	// interrups.
	currentdigit++;
	if (currentdigit>=DIGIT_COUNT) currentdigit=0;
}





//
// Read a 32-bit value from the EEPROM
//
uint32_t eeprom_read_long(uint16_t *loc) {
	uint32_t num;
    num = eeprom_read_word((uint16_t *)(loc+1));
    num <<=16;
    num |= eeprom_read_word((uint16_t *)loc);
    return num;
}


//
// Write a 32-bit value to the EEPROM
//
void eeprom_write_long(uint16_t *loc, uint32_t val) {
	eeprom_write_word(loc, val & 0xffff);
	eeprom_write_word(++loc, val >>16);
}




//
// Initialize and clear all ports to output
//
void InitializePorts() {
	// All available ports are output
	DDRA =0x03;	
	DDRB =0xFF;
	DDRD =0xFF;

	// Output 0 on all ports
	PORTA = 0x00;
	PORTB = 0x00;
	PORTD = 0x00;
}



//
// Start the refresh timer
//
void InitializeTimer() {
	// Initialize display refresh timer
	TIMSK = _BV(TOIE0);   			//Enable TCNT0 overflow
	TCNT0 = 0;         				//Reset TCNT0
	TCCR0A = 0;						//Normal overflow
	TCCR0B = _BV(CS00) |_BV(CS01);  //Prescale /64

	// Globally enable interrupts
	sei();					
}


//
// The main procedure.
//	Inititialize system
//	Get a new random number
//	Fetch the message from flash into a buffer
//	Scroll the message once on the display
//  Shut down the CPU to conserve power
//
int main (void) {
	uint32_t ctx;			// Used for storing the random seed
	char currentmessage[64];// Buffer for the current message
	int len;
	int msgnr;


	InitializePorts();
	InitializeTimer();

	// Wait a  while so the user can get the lid fully opened.
	Delay_ms(1500);

	// Get the old random seed from nonvolatile memory,
	// create a new random number, and then write the new
	// seed back to the nonvolatile memory
	ctx = eeprom_read_long(&eeprom_seed);
	msgnr = random_r(&ctx) % MESSAGE_COUNT;
	eeprom_write_long(&eeprom_seed, ctx);

	// Clear the message buffer
	for (int i=0; i<sizeof(currentmessage); i++) {
		currentmessage[i]=32;
	}

	// Copy the randomly selected message to the message buffer,
	// but start at the 8'th character in the buffer so there are
	// eight spaces in the start of the message to be displayed. This
	// will cause the messsage to be scrolled onto the screen instead
	// of just appearing.
	strcpy_P(&currentmessage[DIGIT_COUNT], 
			 (PGM_P)pgm_read_word(&(messages[msgnr])));
	len=strlen(currentmessage);
	currentmessage[len]=32;

	// Scroll the message onto the display with a speed of 5 characters per
	// second. Do this by taking the ascii values for the 8 characters
	// starting in the messsage buffer offset by the 'ofset' variable.
	// Use the characternap to look up the bitmap of the characters and store
	// the bitmaps in the segs[]-array used by the interrup-handler.
	for (int offset=0; offset<len+1; offset++) {
		for (int i=0; i<DIGIT_COUNT; i++) {
			int p = currentmessage[i+offset]-32;
			segs[i]=pgm_read_word(&(charactermap[p]));
		}
	Delay_ms(200);
	}

	// The message is now displayed.  Just go to a really deep sleep 
	// until the user closes the lid so the power shuts off.

	set_sleep_mode(SLEEP_MODE_PWR_DOWN);
	cli();
	sleep_mode();
}



